from apscheduler.scheduler import Scheduler
import logging
import signal

import bigpanda.agent.core.executors as executors
from bigpanda.agent.plugins.agent.heartbeat.plugin import AgentHeartbeat
from bigpanda.agent.core.client import DataApiClient
from bigpanda.agent.core.governor import Governor

from bigpanda.agent.plugins.nagios.state.plugin import NagiosStatePlugin
from bigpanda.agent.plugins.pagerduty.incidents.plugin import PagerDutyIncidentsPlugin
from bigpanda.agent.plugins.zabbix.api.plugin import ZabbixAPIPlugin
from bigpanda.agent.plugins.zenoss.api.plugin import ZenossAPIPlugin

log = logging.getLogger('agent')

plugins = {"agent/heartbeat" : AgentHeartbeat("agent/heartbeat"),
           "nagios/state": NagiosStatePlugin("nagios/state"),
           "zabbix/api": ZabbixAPIPlugin("zabbix/api"),
           "zenoss/api": ZenossAPIPlugin("zenoss/api"),
           "pagerduty/incidents/outbrain": PagerDutyIncidentsPlugin("pagerduty/incidents/outbrain"),
           "pagerduty/incidents/wix": PagerDutyIncidentsPlugin("pagerduty/incidents/wix"),
           "pagerduty/incidents/elance": PagerDutyIncidentsPlugin("pagerduty/incidents/elance"),
           "pagerduty/incidents/bump": PagerDutyIncidentsPlugin("pagerduty/incidents/bump"),
           "pagerduty/incidents/fiverr": PagerDutyIncidentsPlugin("pagerduty/incidents/fiverr"),
           "pagerduty/incidents/airbnb": PagerDutyIncidentsPlugin("pagerduty/incidents/airbnb"),
           "pagerduty/incidents/quora": PagerDutyIncidentsPlugin("pagerduty/incidents/quora"),
           "pagerduty/incidents/zendesk": PagerDutyIncidentsPlugin("pagerduty/incidents/zendesk"),
           "pagerduty/incidents/onavo": PagerDutyIncidentsPlugin("pagerduty/incidents/onavo"),
           "pagerduty/incidents/jive": PagerDutyIncidentsPlugin("pagerduty/incidents/jive"),
           "pagerduty/incidents/evernote": PagerDutyIncidentsPlugin("pagerduty/incidents/evernote"),
           "pagerduty/incidents/zynga": PagerDutyIncidentsPlugin("pagerduty/incidents/zynga"),
           "pagerduty/incidents/sephora": PagerDutyIncidentsPlugin("pagerduty/incidents/sephora"),
           "pagerduty/incidents/twitchtv": PagerDutyIncidentsPlugin("pagerduty/incidents/twitchtv"),
           "pagerduty/incidents/crittercism": PagerDutyIncidentsPlugin("pagerduty/incidents/crittercism"),
           "pagerduty/incidents/lotame": PagerDutyIncidentsPlugin("pagerduty/incidents/lotame"),
           "pagerduty/incidents/gettaxi": PagerDutyIncidentsPlugin("pagerduty/incidents/gettaxi"),
           "pagerduty/incidents/bigpanda": PagerDutyIncidentsPlugin("pagerduty/incidents/bigpanda")}


scheduler = None
def quit(signum, frame):
    log.debug("Got signal %d" % signum)
    for j in scheduler.get_jobs():
        scheduler.unschedule_job(j)
    scheduler.shutdown(wait=False)

class Agent():
    def  __init__(self, agent_config):
        global scheduler
        self.executors = []

        log.debug("Adding signal handlers.")
        signal.signal(signal.SIGINT, quit)
        signal.signal(signal.SIGTERM, quit)

        scheduler = Scheduler()
        scheduler.configure(coalesce=True, daemonic=False)

        log.info("Initializing Governor")
        governorClient = self._create_client(agent_config)
        self.governor = Governor(scheduler, agent_config, governorClient)
        scheduler.start()

        plugins_configuration = agent_config["plugins"]
        state_dir = agent_config.get("state_dir", '/var/lib/bigpanda/state')

        for plugin in plugins:
            if plugin in plugins_configuration:
                plugin_conf = plugins_configuration.get(plugin)
                if plugin_conf["enabled"]:
                    log.info("Plugin {0} is enabled and will be executed.".format(plugin))
                    plugin_instance = plugins.get(plugin)

                    client = self._create_client(agent_config, plugin_conf)

                    self.executors.append(executors.PythonExecutor(plugin_instance, client, plugin_conf, state_dir))

    def run(self):
        for executor in self.executors:
            job_name = executor.get_name()
            execution_interval = executor.get_execution_interval()
            scheduler.add_interval_job(executor.execute, seconds=execution_interval, name=job_name)

    def _create_client(self, agent_config, plugin_config=None):
        backend_config = agent_config["backend"]
        api_token = backend_config["api_token"]

        if plugin_config and plugin_config.has_key("bigpanda_api_token") and plugin_config["bigpanda_api_token"]:
            api_token = plugin_config["bigpanda_api_token"]

        log.info("Creating client with token: {0}".format(api_token))
        return DataApiClient(backend_config["address"],
                             api_token,
                             backend_config["verify_certificate"],
                             agent_config["agent_id"],
                             agent_config["version"],
                             backend_config.get("timeout", None))
