import logging
import simplejson as json
import os
import shutil
import subprocess
import time
import sys

from bigpanda.agent.core.configuration import hash_file

log = logging.getLogger('agent.commands')


class CommandNotFoundException(Exception):
    def __init__(self, command):
        Exception.__init__(self, "No such command")


class CommandError(Exception):
    pass


class AgentCommands(object):
    def __init__(self, agent_config, scheduler):
        self._forbidden_commands = ['run']
        self._agent_config = agent_config
        self._scheduler = scheduler

    def __contains__(self, item):
        if item.startswith('_') \
                or item not in self.__class__.__dict__ \
                or item in self._forbidden_commands:
            return False
        return True

    def run(self, command, args=None):
        """Run a command, passing any arguments in args. args will be expanded as
        *args or **kwargs according to its type.
        """

        if command not in self:
            raise CommandNotFoundException(command)

        method = getattr(self, command)
        log.info("%s: Running command" % command)
        if isinstance(args, list):
            method(*args)
        elif isinstance(args, dict):
            method(**args)
        else:
            method(args)

        log.info("%s: Command finished" % command)

    def update_config(self, **config):
        """Gets a new config json from the server and updates the config file.
        Will refuse the update our config file was updated since agent startup.
        """

        config_file_path = self._agent_config['config_file_path']
        config_backup_path = self._agent_config['config_backup_path']
        new_file_name = config_file_path + ".new"

        # Override old config file if it hasn't changed since startup
        old_config_hash = hash_file(config_file_path)
        if old_config_hash != self._agent_config['config_hash']:
            raise CommandError("config file changed since startup, rejecting remote update.")

        # Backup old config file, unless we're already using the backup. In this case,
        # don't overwrite it.
        try:
            if not self._agent_config['using_backup']:
                shutil.copy(config_file_path, config_backup_path)
        except Exception as e:
            raise CommandError("can't backup old config file: %s" % e)

        # Write the new config to a temp file
        with open(new_file_name, "w") as new_file:
            json.dump(config, new_file, indent=True)
            new_file.write("\n")

        log.info("updating config file.")
        os.rename(new_file_name, config_file_path)

    def disable_service(self, args):
        """Disables the service. Shuts down and refuses to go back up using a indicator file."""

        config_dir = os.path.dirname(self._agent_config['config_file_path'])
        disable_service = os.path.join(config_dir, 'disable_service')

        log.info("disable_service: creating %s." % disable_service)
        open(disable_service, 'w').close()

        log.info("disable_service: shutting down.")
        self._scheduler.shutdown()

    def restart_service(self, args):
        """Restart the service.
        Unschedule all jobs and reexec ourselves.
        """

        log.info("restart_service: unscheduling jobs.")
        jobs = self._scheduler.get_jobs()
        for j in jobs:
            self._scheduler.unschedule_job(j)

        log.info("restart_service: restarting.")
        args = list()
        if sys.executable != sys.argv[0]:
            args.append(sys.executable)
        args.extend(sys.argv)
        print sys.executable, args
        os.execv(sys.executable, args)


def timed_poll(proc, timeout):
    "Poll a subprocess.Popen instance with a timeout."

    wait_time = 0
    while wait_time < timeout:
        status = proc.poll()
        if status is not None:
            return status
        time.sleep(1)
        wait_time += 1
