from bigpanda.agent.core.spi.exceptions import PluginExecutionException
from bigpanda.agent.core.client import *
import simplejson as json
import traceback
import os
import copy

log = logging.getLogger('agent')

class PythonExecutor(object):
    def __init__(self, plugin_instance, client, config, state_dir):
        self.plugin_instance = plugin_instance
        self.config = config["config"]

        self.state_dir = state_dir
        self.client = client
        self.endpoint = config["endpoint"]
        self.action = config["action"]
        self.execution_interval = config.get("execution_interval", 10)
        self.state = None

        max_retry_time_minutes = config.get("max_retry_minutes", 180)
        self.max_retries = max_retry_time_minutes * 60 / self.execution_interval
        self.send_attempt = 0

        self.try_load_state()


    def _handle_exception(self, message, **kwargs):
        log.exception(message)
        self.client.sendErrors([dict(message=message, traceback=traceback.format_exc(2), data=kwargs)])

    def execute(self):
        try:
            try:
                done = False

                # Make sure the plugins cannot change the executor state
                state = copy.deepcopy(self.state)

                while not done:
                    result, state, done, errors = self.plugin_instance.execute(self.config, state)

                    if self.send_attempt >= self.max_retries:
                        self.send_attempt = 0

                        message = "Max Attempts for plugin {0} reached. ignoring batch".format(self.get_name())
                        log.warn(message)

                        self.client.sendErrors([dict(message=message)])

                    else:
                        if result:
                            log.debug("Sending results for plugin {0} to backend server".format(self.get_name()))

                            self.send_attempt += 1
                            self.client.send(self.endpoint, self.action, result)

                            self.send_attempt = 0
                            log.debug("Done sending results for plugin {0} to backend server".format(self.get_name()))

                        if errors:
                            log.debug("Sending errors for plugin {0} to backend server".format(self.get_name()))
                            self.client.sendErrors(errors)
                            log.debug("Done sending errors for plugin {0} to backend server".format(self.get_name()))


                    self.state = state
                    self.try_persist_state()

            except InvalidActionExcetion:
                self._handle_exception("Invalid HTTP action used, attempt {0} out of {1}".
                                       format(self.send_attempt, self.max_retries),
                                       endpoint=self.endpoint,
                                       action=self.action)
            except ClientException:
                self._handle_exception("Failed to send data to backend server, attempt {0} out of {1}".
                                       format(self.send_attempt, self.max_retries),
                                       endpoint=self.endpoint,
                                       action=self.action)

            except APIServerException:
                self._handle_exception("Unexpected error in API, attempt {0} out of {1}".
                                       format(self.send_attempt, self.max_retries),
                                       endpoint=self.endpoint,
                                       action=self.action)

            except PluginExecutionException:
                self._handle_exception("Failed to execute plugin, attempt {0} out of {1}".
                                       format(self.send_attempt, self.max_retries),
                                       endpoint=self.endpoint,
                                       action=self.action)
        except ErrorsTransmissionException:
            log.exception("Failed to transmit errors to backend server, attempt {0} out of {1}".
                          format(self.send_attempt, self.max_retries))

    def get_name(self):
        return self.plugin_instance.get_name()

    def get_execution_interval(self):
        return self.execution_interval

    def get_os_safe_name(self):
        return self.get_name().replace('/', '.')

    def get_state_file_path(self):
        return os.path.join(self.state_dir, '{0}.state'.format(self.get_os_safe_name()))

    def try_persist_state(self):
        try:
            if not os.path.exists(self.state_dir):
                os.makedirs(self.state_dir)


            state_json = json.dumps(self.state)
            with open(self.get_state_file_path(), 'w') as f:
                f.write(state_json)

        except Exception:
            log.exception("Couldn't persist plugin state ({0})".format(self.get_name()))

    def try_load_state(self):
        try:
            if not os.path.exists(self.get_state_file_path()):
                log.debug("No plugin state file ({0})".format(self.get_state_file_path()))
                return

            with open(self.get_state_file_path(), 'r') as f:
                state_json = f.read()
                self.state = json.loads(state_json)

        except Exception:
            log.exception("Couldn't load plugin state from file ({0})".format(self.get_state_file_path()))