import copy
import simplejson as json

# THIS IMPORT IS A MUST FOR CORRECTLY INITIALIZING THE LOGGER
import bigpanda.agent as root

import logging
log = logging.getLogger('migration')


class MigrationException(Exception):
    def __init__(self, message, ):
        # Call the base class constructor with the parameters it needs
        Exception.__init__(self, message)


class Migrator:
    def migrate(self, new_version, old_config):
        new_config = copy.deepcopy(old_config)
        root_config = new_config["root"]
        root_config["version"] = new_version
        self._migrate(root_config)
        return new_config


class DefaultMigration(Migrator):
    def _migrate(self, root_config):
        pass


class Migrate_0_0_0(Migrator):
    def _migrate(self, root_config):
        root_config["plugins"]["nagios/state"]["endpoint"] = "/data/events/alerts"


class Migrate_0_9_0(Migrator):
    def _migrate(self, root_config):
        root_config["plugins"]["zabbix/api"] = {
            "enabled": False,
            "endpoint": "/data/events/alerts",
            "action": "POST",
            "config": {
                "host": "http://localhost/zabbix",
                "user": "bigpanda",
                "password": root_config["backend"]["api_token"]
            }
          }


class Migrate_1_2_2(Migrator):
    def _migrate(self, root_config):
        root_config["plugins"]["nagios/state"]["config"]["objects_cache_filename"] = "/usr/local/nagios/var/objects.cache"
        if "timeout" not in root_config["backend"]:
            root_config["backend"]["timeout"] = 30

class Migrate_1_6_0(Migrator):
    def _migrate(self, root_config):
        root_config["plugins"]["zenoss/api"] = {
        "enabled": False,
        "endpoint": "/data/events/alerts",
        "action": "POST",
        "config": {
            "host": "http://localhost:8080",
            "user": "<username>",
            "password": "<password>"
        }
      }

class Migrate_1_6_2(Migrator):
    def _migrate(self, root_config):
        if 'log_dir' not in root_config:
            root_config['log_dir'] = '/var/log/bigpanda/'

class MigrationEngine:
    MIGRATIONS = {"0.0.0": ("0.5.0", Migrate_0_0_0()),
                  "0.5.0": ("0.6.1", DefaultMigration()),
                  "0.6.1": ("0.6.2", DefaultMigration()),
                  "0.6.2": ("0.6.3", DefaultMigration()),
                  "0.6.3": ("0.8.0", DefaultMigration()),
                  "0.8.0": ("0.9.0", Migrate_0_9_0()),
                  "0.9.0": ("0.9.2", DefaultMigration()),
                  "0.9.2": ("0.9.3", DefaultMigration()),
                  "0.9.3": ("0.9.4", DefaultMigration()),
                  "0.9.4": ("0.11.0", DefaultMigration()),
                  "0.11.0": ("0.11.1", DefaultMigration()),
                  "0.11.1": ("0.11.2", DefaultMigration()),
                  "0.11.2": ("0.11.3", DefaultMigration()),
                  "0.11.3": ("1.1.0", DefaultMigration()),
                  "1.1.0": ("1.1.1", DefaultMigration()),
                  "1.1.1": ("1.2.0", DefaultMigration()),
                  "1.2.0": ("1.2.1", DefaultMigration()),
                  "1.2.1": ("1.2.2", Migrate_1_2_2()),
                  "1.2.2": ("1.4.0", DefaultMigration()),
                  "1.4.0": ("1.5.0", DefaultMigration()),
                  "1.5.0": ("1.6.0", Migrate_1_6_0()),
                  "1.6.0": ("1.6.2", DefaultMigration()),
                  "1.6.2": ("1.6.3", Migrate_1_6_2()),
                  }

    def read_conf(self):
        from bigpanda.agent.core.configuration import JsonConfigurationParser, ConfigurationParseException

        config_file_path, config_backup_path = root.get_config_filpath(root.get_env())
        try:
            configurator = JsonConfigurationParser(config_file_path)
            return configurator.parse()
        except ConfigurationParseException:
            raise MigrationException("Could not migrate configuration: failed to parse configuration file")

    def get_version(self, config):
        return config["root"].get("version") or "0.0.0"

    def find_migration(self, config):
        version = self.get_version(config)
        return self.MIGRATIONS.get(version)

    def run_migrations(self, current_config=None, update_file=True):
        if current_config is None:
            current_config = self.read_conf()
        migration = self.find_migration(current_config)

        while migration:
            new_version = migration[0]
            migration_logic = migration[1]

            current_config = migration_logic.migrate(new_version, current_config)
            migration = self.find_migration(current_config)

        if update_file:
            config_file_path, config_backup_path = root.get_config_filpath(root.get_env())
            with open(config_file_path, "w") as fp:
                json.dump(current_config, fp, indent=2 * ' ')

        return current_config

