import re
import os
from bigpanda.agent.core.spi.plugins import Plugin
from bigpanda.agent.core.spi.exceptions import PluginExecutionException
from bigpanda.agent.utils.file_utils import Tail
from bigpanda.agent.plugins.nagios.state.cache import ObjectCacheParser, ObjectCache
from collections import namedtuple
import logging
import socket
import datetime
import time

log = logging.getLogger('agent.plugins.nagios.state')


class NagiosStatePlugin(Plugin):

    EVENT_TYPES_MAPPING = {
        "Alert" 	: 	["HOST NOTIFICATION", "SERVICE NOTIFICATION", "SERVICE ALERT", "SERVICE FLAPPING ALERT", "HOST FLAPPING ALERT", "HOST ALERT", "ACKNOWLEDGE_SVC_PROBLEM", "ACKNOWLEDGE_HOST_PROBLEM", "CURRENT SERVICE STATE", "CURRENT HOST STATE"],
        "Downtime"	:   ["HOST DOWNTIME ALERT", "SERVICE DOWNTIME ALERT", "SCHEDULE_SVC_DOWNTIME", "SCHEDULE_HOST_DOWNTIME"]
    }

    HOST_CHECKS = ["HOST FLAPPING ALERT", "HOST ALERT", "CURRENT HOST STATE", "ACKNOWLEDGE_HOST_PROBLEM"]

    STATUS_MAPPING = {
        "OkSuspect"			:   [("OK", "SOFT"), ("UP", "SOFT")],
        "Ok"				:   [("OK", "HARD"), ("UP", "HARD")],
        "WarningSuspect"	:   [("WARNING", "SOFT"), ("UNKNOWN", "SOFT")],
        "Warning"			: 	[("WARNING", "HARD"), ("UNKNOWN", "HARD")],
        "CriticalSuspect"	:   [("CRITICAL", "SOFT"), ("DOWN", "SOFT"), ("UNREACHABLE", "SOFT")],
        "Critical"			: 	[("CRITICAL", "HARD"), ("DOWN", "HARD"), ("UNREACHABLE", "HARD")],
    }

    DOWNTIME_STATUS_MAPPING = {
        "Started"			:   ["STARTED"],
        "Stopped"			:   ["STOPPED"]
    }

    PARSERS = {
        "CURRENT HOST STATE"	:   namedtuple("CurrentHostStateEvent", 	["host", "status", "soft_or_hard", "retries", "desc"]),
        "CURRENT SERVICE STATE"	:   namedtuple("CurrentServiceStateEvent", 	["host", "check_name", "status", "soft_or_hard", "retries", "desc"]),

        "HOST ALERT"			:   namedtuple("HostAlertEvent", 			["host", "status", "soft_or_hard", "retries", "desc"]),
        "SERVICE ALERT"			:   namedtuple("ServiceAlertEvent", 		["host", "check_name", "status", "soft_or_hard", "retries", "desc"]),

        "HOST NOTIFICATION" 	: 	namedtuple("HostNotificationEvent", 	["contact", "host", "status", "notification_type", "desc"]),
        "HOST FLAPPING ALERT"	:   namedtuple("HostFlappingAlert",         ["host", "status", "desc"]),

        "PASSIVE SERVICE CHECK"	:   namedtuple("PassiveServiceCheckEvent", 	["host", "check_name", "retries", "desc"]),
        "SERVICE NOTIFICATION" 	: 	namedtuple("ServiceNotificationEvent", 	["contact", "host", "check_name", "status", "notification_type", "desc"]),
        "SERVICE FLAPPING ALERT":   namedtuple("ServiceFlappingEvent", 		["host", "check_name", "status", "desc"]),

        "ACKNOWLEDGE_SVC_PROBLEM":  namedtuple("AcknowledgeServiceProblem", ["host", "check_name", "sticky", "notify", "persistent_comment", "user", "desc"]),
        "ACKNOWLEDGE_HOST_PROBLEM": namedtuple("AcknowledgeHostProblem",    ["host", "sticky", "notify", "persistent_comment", "user", "desc"]),

        #"HOST DOWNTIME ALERT"	: 	namedtuple("HostDowntimeEvent", 		["host", "status", "desc"]),
        #"SERVICE DOWNTIME ALERT": 	namedtuple("ServiceDowntimeEvent", 		["host", "check_name", "status", "desc"]),

        #"SCHEDULE_SVC_DOWNTIME": namedtuple("ScheduleServiceDowntime", ["host", "check_name", "start_timestamp", "end_timestamp", "fixed_or_flexible", "triggered_by", "flexible_duration", "user", "desc"]),
        #"SCHEDULE_HOST_DOWNTIME": namedtuple("ScheduleHostDowntime", ["host", "start_timestamp", "end_timestamp", "fixed_or_flexible", "triggered_by", "flexible_duration", "user", "desc"])
    }

    # System events which are internal to Nagios and are of no aperrent use for us
    NAGIOS_SYSTEM_EVENTS = ["LOG ROTATION", "LOG VERSION", "TIMEPERIOD TRANSITION", "SERVICE EVENT HANDLER"]
    # Events that we decided to filter out
    EVENTS_TO_SKIP = ["INITIAL SERVICE STATE", "INITIAL HOST STATE", "PASSIVE SERVICE CHECK",
                      "HOST NOTIFICATION", "SERVICE NOTIFICATION"]
    # External commands that we decided to filter out
    EXTERNAL_COMMANDS_TO_SKIP = ["SCHEDULE_HOST_SVC_DOWNTIME", "SCHEDULE_FORCED_SVC_CHECK", "DISABLE_SVC_NOTIFICATIONS", "DEL_SVC_DOWNTIME", "PROCESS_SERVICE_CHECK_RESULT"]

    EVENT_TYPES_TO_EXCLUDE = NAGIOS_SYSTEM_EVENTS + EVENTS_TO_SKIP + EXTERNAL_COMMANDS_TO_SKIP

    def _init(self):
        self.event_regex = re.compile('^\[(\d+)\] ([A-Z,\s]+): (.*)$')
        self.command_regex = re.compile('^\[(\d+)\] EXTERNAL COMMAND: ([A-Z,\s,_]+);(.*)$')
        self.hostname = socket.getfqdn()
        self.object_cache = None

    def _process(self, tail, min_timestamp):
        events = []

        for line in tail.process():
            if line == tail.EOF:
                return line
            else:
                event = self.parse_line(line, min_timestamp)
                if event:
                    events.append(event)

        return events

    def _create_tail(self, config, tail_state=None):
        log_file = config['nagios_log_file']
        log.debug("Parsing " + log_file)
        if os.path.exists(log_file):

            tail_state = tail_state or (0, None, 0)
            tail = Tail(log_file, 1048576, *tail_state)

            return tail
        else:
            raise PluginExecutionException("Nagios log file does not exist in: {0}. "
                                           "Nagios alerts will not be processed!".format(log_file))

    def _refresh_object_cache(self, objects_cache_filename, state):
        log.debug('Refreshing objects cache')
        parser = ObjectCacheParser()
        object_cache = parser.parse(objects_cache_filename)
        state['objects_cache_last_refresh'] = time.mktime(datetime.datetime.now().timetuple())
        log.debug('Successfully refreshed objects cache')
        return object_cache

    def _should_refresh_object_cache(self, config, state):
        if not config.get('objects_cache_filename'):
            return False

        if not self.object_cache:
            return True

        interval = config.get('objects_cache_refresh_interval', 10 * 60) # 10 minutes by default
        refresh_watermark = datetime.datetime.now() - datetime.timedelta(seconds=interval)
        refresh_watermark_posix = time.mktime(refresh_watermark.timetuple())
        last_refresh_posix = state.get('objects_cache_last_refresh')
        return not last_refresh_posix or refresh_watermark_posix > last_refresh_posix



    def _execute(self, config, state):
        events = []

        done = False

        if not state.get('tail_state'):
            tail = self._create_tail(config)
        else:
            tail = self._create_tail(config, state.get('tail_state'))

        try:
            if self._should_refresh_object_cache(config, state):
                self.object_cache = self._refresh_object_cache(config.get('objects_cache_filename'), state)
        except Exception:
            log.exception('Failed to refresh objects cache')

        if not self.object_cache:
            self.object_cache = ObjectCache()

        # Earliest unix-time to collect events
        history_days = config.get("history_days", 7)
        min_datetime = datetime.datetime.now() - datetime.timedelta(days=history_days)
        min_timestamp = int(time.mktime(min_datetime.timetuple()))

        result = self._process(tail, min_timestamp)

        state['tail_state'] = tail.get_state()

        if result == tail.EOF:
            done = True
        else:
            events = result

        if not len(events):
            log.debug("Didn't find any events")
            return None, state, done
        else:
            log.debug("Found {0} events".format(len(events)))
            return {"alerts": events}, state, done

    def parse_event(self, line, match, min_timestamp):
        timestamp, event_type, fields = match.groups()
        if not self._skip_event_type(event_type):
            timestamp = int(timestamp)

            if timestamp < min_timestamp:
                return None

            normalized_event_type = self._normalize_event_type(event_type)
            if not normalized_event_type:
                log.warning("Unknown event type: {0}".format(event_type))
                return None

            parser = self.PARSERS.get(event_type)
            if not parser:
                log.warning("No matching parser for event type: [{0}]. Original line: {1}".format(event_type, line))
                return None

            try:
                parsed_event = parser._make(fields.split(";"))
            except Exception:
                message = "Failed to parse event line of type [{0}]. Original line: {1}".format(event_type, line)
                log.error(message)
                self._push_error(message, eventType=event_type, line=line)
                return None

            # Mandatory fields
            event = {}

            event["timestamp"] = timestamp
            event["description"] = parsed_event.desc

            event["source_system"] = "nagios"
            event["primary_property"] = "_check"
            event["properties"] = {}

            host = self._normalize_host(parsed_event.host)
            event["properties"]["_host"] = host
            hostgroups = self.object_cache.get_hostgroups(host)
            if hostgroups and len(hostgroups) > 0:
                event["properties"]["hostgroups"] = hostgroups

            event["incident_key"] = self._normalize_host(parsed_event.host)

            normalized_status = self._normalize_status(event_type, parsed_event)
            if not normalized_status:
                log.warn("Failed to normalize status. Setting status to 'Unknown'. original line: {0}".format(line))
                event["status"] = 'Unknown'
                event["properties"]["original_status"] = parsed_event.status
                event["properties"]["original_event_type"] = event_type
            else:
                event["status"] = normalized_status


            if self._entails_assignee(event_type):
                assignee = parsed_event.user.strip()
                if assignee:
                    event["properties"]["_assignee"] = assignee

            check_name = self._normalize_check_name(parsed_event, event_type)
            if check_name:
                event["properties"]["_check"] = check_name
                event["incident_key"] += "__" + check_name
                servicegroups = self.object_cache.get_servicegroups(host, check_name)
                if servicegroups and len(servicegroups) > 0:
                    event["properties"]["servicegroups"] = servicegroups


            return event

    def parse_line(self, line, min_timestamp):
        # First check if external command because the event regexp matches against external commands lines
        match = self.command_regex.match(line)
        if match:
            return self.parse_event(line, match, min_timestamp)

        match = self.event_regex.match(line)
        if match:
            return self.parse_event(line, match, min_timestamp)

    def _host_check(self, event_type):
        return event_type in self.HOST_CHECKS

    def _normalize_check_name(self, parsed_event, event_type):
        if "check_name" in parsed_event._fields:
            return parsed_event.check_name
        if self._host_check(event_type):
            return "__ping"
        return "__unknown_check"

    def _normalize_event_type(self, current_event_type):
        for evt_type in self.EVENT_TYPES_MAPPING:
            if current_event_type in self.EVENT_TYPES_MAPPING.get(evt_type):
                return evt_type

    def _normalize_status(self, event_type, parsed_event):
        if event_type in ["ACKNOWLEDGE_SVC_PROBLEM", "ACKNOWLEDGE_HOST_PROBLEM"]:
            return "Acknowledged"
        elif event_type in self.EVENT_TYPES_MAPPING["Downtime"]:
            if "SCHEDULE" in event_type:
                return "Scheduled"
            else:
                original_status = parsed_event.status
                for status in self.DOWNTIME_STATUS_MAPPING:
                    if original_status in self.DOWNTIME_STATUS_MAPPING.get(status):
                        return status
        else:
            if "soft_or_hard" in parsed_event._fields:
                soft_or_hard = parsed_event.soft_or_hard
            else:
                soft_or_hard = "HARD"

            original_status = parsed_event.status
            for status in self.STATUS_MAPPING:
                if (original_status, soft_or_hard) in self.STATUS_MAPPING.get(status):
                    return status

    def _normalize_host(self, host):
        if host == 'localhost':
            host = self.hostname
        return host

    def _entails_status(self, event_type, parsed_event):
        return "status" in parsed_event._fields\
               or event_type in ["ACKNOWLEDGE_SVC_PROBLEM", "ACKNOWLEDGE_HOST_PROBLEM", "SCHEDULE_SVC_DOWNTIME", "SCHEDULE_HOST_DOWNTIME"]

    def _entails_assignee(self, event_type):
        return event_type in ["ACKNOWLEDGE_SVC_PROBLEM", "ACKNOWLEDGE_HOST_PROBLEM", "SCHEDULE_SVC_DOWNTIME", "SCHEDULE_HOST_DOWNTIME"]

    def _skip_event_type(self, event_type):
        return event_type in self.EVENT_TYPES_TO_EXCLUDE