from bigpanda.agent.core.spi.plugins import Plugin
from bigpanda.agent.core.spi.exceptions import PluginExecutionException
import logging
import requests
import datetime
import calendar
import json

log = logging.getLogger('agent.plugins.pagerduty.incidents')


class PagerDutyIncidentsPlugin(Plugin):

    URL = "https://<subdomain>.pagerduty.com/api/v1/log_entries"

    STATUS_MAPPING = {
            "triggered"     : "Critical",
            "trigger"       : "Critical",
            "acknowledged"  : "Acknowledged",
            "acknowledge"   : "Acknowledged",
            "resolved"      : "Ok",
            "resolve"       : "Ok",
            "unacknowledge" : "Critical",
            "assign"        : "Critical",
            "escalate"      : "Critical"
        }

    def _init(self):
        self.set_retriever(PagerDutyClient())

    def _init_state(self):
        state = {}
        state['last_timestamp'] = 0

        return state

    def set_retriever(self, retriever):
        self.retriever = retriever

    def _execute(self, config, state):

        done = False
        min_datetime = None
        process_events = []
        history_days = config.get("history_days", 7)
        offset = 0
        last_log_time = None

        if state.has_key('last_timestamp') and state["last_timestamp"]:
            min_datetime = state['last_timestamp']
        else:
            min_datetime = (datetime.datetime.utcnow() - datetime.timedelta(days=history_days)).isoformat()

        while not done:
            log.debug("Querying from offset: {0}".format(offset))

            valid, data = self._request_events(config, min_datetime, offset)

            if not valid:
                subdomain = config.get("subdomain")
                raise PluginExecutionException("Invalid response for domain {0}: {1}".format(subdomain, data))

            else:
                log_entries = data.get('log_entries')
                sorted_events = self._sort_events(log_entries)

                # Add the new events before the available ones
                # This is because the API of pagerduty returns the events from last to first
                process_events = self._process(sorted_events) + process_events

                log.debug("Position: [requested_offset: {0}, retured_offset: {1}, limit: {2}, total: {3}]". \
                          format(offset,
                                 data.get('offset'),
                                 data.get('limit'),
                                 data.get('total')))

                next = data.get('offset') + data.get('limit')

                if data.get('total') > next:
                    offset = next

                    last_log_time = self._get_last_log_time(log_entries, last_log_time)
                else:
                    done = True
                    last_log_time = self._get_last_log_time(log_entries, last_log_time)

                    # Only change if last event available, if no event - means we need to retry again next time
                    if last_log_time:
                        state['last_timestamp'] = last_log_time

        if not len(process_events):
            log.debug("Didn't find any events")
            return None, state, True
        else:
            log.debug("Found {0} events".format(len(process_events)))
            return {"alerts": process_events, "namespace": config.get("agent_id")}, state, True

    def _process(self, events):
        parsed_events = []

        for event in events:

            if self._is_status_changed(event):
                parsed = self._parse_incident(event["incident"])
                self._update_incident(event, parsed)

                parsed_events.append(parsed)

        return parsed_events

    def _request_events(self, config, min_date, offset=0):

        subdomain = config.get("subdomain")
        token = config.get("token")

        request_url = self.URL.replace("<subdomain>", subdomain)

        headers = {'Content-Type': 'application/json', "Authorization": "Token token=" + token}
        data = {"since": min_date, 'offset': offset, 'include[]': 'incident'}

        log.debug("Executing: {0}, args: {1}".format(request_url, data))

        r = self.retriever.get(request_url, data, headers)

        if r.status_code == 200:
            return True, json.loads(r.text)
        else:
            return False, r.text

    def _sort_events(self, events):
        return reversed(events)

    def _parse_incident(self, incident):
        parsed = {}

        parsed["incident_key"] = incident["id"]
        parsed["status"] = self._normalize_status(incident)

        parsed["source_system"] = "pagerduty"
        parsed["primary_property"] = "service"

        parsed["properties"] = {}
        parsed["properties"]["_check"] = self._normalize_check(incident)
        server = self._normalize_server(incident)
        if server:
            parsed["properties"]["_host"] = server
        parsed["properties"]["service"] = incident["service"]["name"]

        parsed["links"] = []
        parsed["links"].append({"label": "PagerDuty Incident", "url": incident["html_url"]})

        return parsed

    def _update_incident(self, event, parsed):
        # Always use the event time - incident may already been updated with new data

        parsed["timestamp"] = self._parse_time(event["created_at"])
        parsed["status"] = self._normalize_status_by_type(event)

        assignee = self._normalize_assigned_owner(event)

        if assignee:
            parsed["properties"]["_assignee"] = assignee

        description = self._normalize_event_description(event, parsed)

        if description:
            parsed["description"] = description

        if self._key_avail(event, "channel"):
            parsed["properties"]["initiator"] = event["channel"]["type"]

    def _is_status_changed(self, event):
        event_type = event.get('type')

        return event_type == "acknowledge" or \
               event_type == "trigger" or \
               event_type == "unacknowledge" or \
               event_type == "assign" or \
               event_type == "escalate" or \
               event_type == "resolve"

    def _parse_time(self, time_value):
        last_changed = datetime.datetime.strptime(time_value, "%Y-%m-%dT%H:%M:%SZ")
        return int(calendar.timegm(last_changed.timetuple())) # Calendar assume the time is in UTC

    def _normalize_status(self, incident):
        status = incident.get('status')

        if self.STATUS_MAPPING.has_key(status):
            return self.STATUS_MAPPING[status]
        else:
            return "unknown"
            # Throw exception

    def _normalize_status_by_type(self, event):
        status = event.get('type')

        if self.STATUS_MAPPING.has_key(status):
            return self.STATUS_MAPPING[status]
        else:
            return "unknown"

    def _normalize_check(self, incident):
        data = incident["trigger_summary_data"]

        if data and self._key_avail(data, "SERVICEDESC"):
            return data["SERVICEDESC"]

        elif data and self._key_avail(data, "subject"):
            return data["subject"]

        elif self._key_avail(incident, "incident_key"):
            return incident["incident_key"]

        else:
            return None

    def _normalize_server(self, incident):
        data = incident["trigger_summary_data"]

        if data and data.has_key("HOSTNAME"):
            return data["HOSTNAME"]
        else:
            return None

    def _normalize_incident_owner(self, incident):
        return self._normalize_user(incident, "assigned_to_user")

    def _normalize_assigned_owner(self, event):
        return self._normalize_user(event, "assigned_user")

    def _normalize_user(self, item, key):
        if self._key_avail(item, key):
            return "{0} [{1}]".format(item[key]["name"], item[key]["email"])
        else:
            return None

    def _normalize_event_description(self, event, parsed):
        description = ""

        # Add description for what happened 
        if event["type"] == "assign":
            description = "Assigned to: " + parsed["properties"]["_assignee"]
        elif event["type"] == "unacknowledge":
            description = "Unacknowledged"
        elif event["type"] == "acknowledge":
            description = "Acknowledged"

            if event.has_key("agent") and event["agent"].has_key("name"):
                description += " by " + event["agent"]["name"]

        if self._key_avail(event, "note"):
            description += event["note"]
        elif self._key_avail(event, "channel") and self._key_avail(event["channel"], "summary"):
            description += event["channel"]["summary"]

        if description == "":
            description = None

        return description

    def _get_last_log_time(self, log_entries, last_log):
        # Logs are ordered by latest first.
        # If last log already set, it is the latest
        if last_log:
            return last_log

        if log_entries and len(log_entries) > 0:
            return log_entries[0]["created_at"]
        else:
            return None

    def _key_avail(self, item, key):
        return item.has_key(key) and item[key]



class PagerDutyClient(object):

    def get(self, request_url, data, headers):
        return requests.get(request_url, params=data, headers=headers, verify=True)