from bigpanda.agent.core.spi.plugins import Plugin
from bigpanda.agent.core.spi.exceptions import PluginExecutionException
import logging
import socket
import pyzabbix
import datetime
import time

log = logging.getLogger('agent.plugins.zabbix.api')

class ZabbixAPIPlugin(Plugin):

    STATUS_MAPPING = {
        "Warning"			:   "WARNING",
        "Average"			:   "WARNING",
        "High"	    		:   "CRITICAL",
        "Disaster"			:   "CRITICAL",
    }

    STATUS_TRIGGER_MAPPING = {
        "2"			:   "WARNING",
        "3"			:   "WARNING",
        "4"    		:   "CRITICAL",
        "5"			:   "CRITICAL",
    }

    TRANSLATE_PRIORITY = {
        "2"			:   "Warning",
        "3"			:   "Average",
        "4"    		:   "High",
        "5"			:   "Disaster",
    }

    UNKNOWN_STATUS = "UNKNOWN"

    BIGPANDA_ACTION_NAME = "BigPanda Action"

    WARN_DESCRIPTION_LENGTH = 10 * 1000

    RETRIEVE_MAX_LIMIT = 500

    def _init(self):
        self.hostname = socket.getfqdn()
        return 0

    def _execute(self, config, state):
        zapi = pyzabbix.ZabbixAPI(config.get('host', None))
        zapi.login(config['user'], config['password'])

        self._validate_actionid(zapi, state)

        min_full_update = self._calculate_reload_interval(config)
        is_first_time = not "last_full_update" in state

        if is_first_time or state["last_full_update"] < min_full_update:
            events = self._update_for_open_items(zapi, config, is_first_time)
            state["last_full_update"] = int(time.mktime(datetime.datetime.now().timetuple()))

        else:
            actionid = config.get("actionid", state.get("actionid"))
            min_load_time = self._calculate_min_historical_timestamp(config)
            zabbix_alerts = self._retrieve_alerts(zapi, actionid, state.get("last_alert_time", min_load_time))
            log.info("{0} alerts found".format(len(zabbix_alerts)))
            events = self._process(zabbix_alerts)

            if len(events) > 0:
                state["last_alert_time"] = int(events[-1]["timestamp"]) - 1

        if not len(events):
            log.debug("Didn't find any events")
            return None, state, True
        else:
            log.debug("Found {0} events".format(len(events)))
            return {"alerts": events}, state, len(events) < self.RETRIEVE_MAX_LIMIT


    def _validate_actionid(self, zapi, state):
        if not "actionid" in state:
            action_id = self.load_action_id(zapi)

            if not action_id:
                log.error("Cannot find BigPanda action! - Stopping execution")
                return None, state, True
            else:
                state["actionid"] = action_id
                log.debug("Found BigPanda notification {0}".format(action_id))

    def _retrieve_alerts(self, zapi, actionid, last_alert_time):

        return zapi.alert.get(
            output='extend',
            actionids=actionid,
            time_from=last_alert_time,
            sortfield='alertid',
            sortorder='ASC',
            limit=self.RETRIEVE_MAX_LIMIT
        )

    def _update_for_open_items(self, zapi, config, first_time):
        if first_time:
            log.info("Update open items for the first time")
        else:
            log.info("Updating open items")

        events = []

        triggers = self._retrieve_open_triggers(zapi, config)
        log.info("{0} triggers found".format(len(triggers)))

        for trigger in triggers:
            event = self.process_open_trigger(trigger, first_time)
            if event:
                events.append(event)

        return events

    def _retrieve_open_triggers(self, zapi, config):
        return zapi.trigger.get(only_true=1,
            skipDependent=1,
            monitored=1,
            active=1,
            output='extend',
            expandDescription=1,
            expandData='hostname',
            lastChangeSince=self._calculate_min_historical_timestamp(config),
            selectItems='extend',
            sortfield='lastchange',
            sortorder='ASC',
            limit=self.RETRIEVE_MAX_LIMIT
        )

    def _process(self, zabbix_events):
        events = []

        for zabbix_event in zabbix_events:
            event = self.process_alert(zabbix_event)
            if event:
                events.append(event)

        return events

    def parse_event_body(self, alertid, body):
        invalidKey = "Invalid"
        data = dict((line.strip().split(':', 1) if ':' in line else [invalidKey, line]) for line in body.strip().split('\n'))

        if invalidKey in data:
            log.warn("Cannot parse message! alert id: {0}, Invalid key len {1}, Message ken {2} message start: {3}, key start: {4}".
                     format(alertid, len(data[invalidKey]), len(body), body[:100], data[invalidKey][:100]))
            return None


        return data

    def process_alert(self, zabbix_alert):
        event = {}

        if not zabbix_alert.get("message") or len(zabbix_alert["message"]) == 0:
            log.error("Received message with no body - ignoring ")
            return None

        alert_body = self.parse_event_body(zabbix_alert["alertid"], zabbix_alert["message"])

        if not alert_body:
            log.warn("No body returned - ignoring alert")
            return None

        event["status"] = self._normalize_alert_status(alert_body.get("severity", -1), alert_body.get("value", 1))

        if not event["status"]:
            log.warn("Received alert with no / unrecognized status - {0}".format(zabbix_alert["message"]))
            return None

        event["timestamp"] = int(zabbix_alert["clock"])

        event["description"] = alert_body.get("description", None)

        event["source_system"] = "zabbix"
        event["primary_property"] = "_check"

        event["properties"] = {}
        event["properties"]["_host"] = self._normalize_host(alert_body.get("hostname", "NO Server"))
        event["properties"]["_check"] = alert_body.get("check", "NO Item")

        event["incident_key"] = event["properties"]["_host"] + "__" + event["properties"]["_check"]

        return event

    def process_open_trigger(self, trigger, first_time):
        event = {}

        event["status"] = self._normalize_trigger_status(trigger.get("priority", -1), trigger.get("value", 1))

        if not event["status"]:
            return None

        # This is an open trigger,
        # In case this is the first time, just use its last change.
        # Else, use now as the trigger timestamp - this will update its time to keep to now.
        if first_time:
            event["timestamp"] = int(trigger["lastchange"])
        else:
            event["timestamp"] = int(time.mktime(datetime.datetime.now().timetuple()))

        event["description"] = trigger.get("description", None)

        event["source_system"] = "zabbix"
        event["primary_property"] = "_check"

        event["properties"] = {}
        event["properties"]["_host"] = self._normalize_host(trigger.get("hostname", "NO Server"))

        if self.has_property(trigger, "items"):
            event["properties"]["_check"] = trigger["items"][0]["name"]

        if event["description"] and len(event["description"]) > self.WARN_DESCRIPTION_LENGTH:
            log.warn("Very Long trigger description!. triggerid: {0}, len: {1}, start: {2}".
                     format(trigger["triggerid"], len(event["description"]), event["description"][:100]))

        event["incident_key"] = event["properties"]["_host"] + "__" + event["properties"]["_check"]

        return event

    def load_action_id(self, zapi):
        actions = zapi.action.get(output=["name"])

        for action in actions:
            if action["name"] == self.BIGPANDA_ACTION_NAME:
                log.info("Found action id: {0}".format(action["actionid"]))
                return action["actionid"]

        raise PluginExecutionException("Cannot find BigPanda Notification action. "
                                       "Please make sure you've configured BigPanda Before running the plugin")

    def has_property(self, item, property):
        return property in item and len(item[property]) > 0

    def _normalize_host(self, host):
        if host == 'localhost':
            host = self.hostname
        return host

    def _normalize_alert_status(self, priority, value):
        if value == "0":
            return "OK"

        status = self.STATUS_MAPPING.get(priority, None)

        if not status:
            log.warn("Ignoring unrecognized status for: priority: {0}, value: {1}".format(priority, value))

        return status

    def _normalize_trigger_status(self, priority, value):
        if value == "0":
            return "OK"

        status = self.STATUS_TRIGGER_MAPPING.get(priority, None)

        if not status:
            log.warn("Ignoring unrecognized status for: priority: {0}, value: {1}".format(priority, value))

        return status

    def _calculate_reload_interval(self, config):
        reload_interval = config.get("reload", {"days": 1})
        min_full_update_date = datetime.datetime.now() - datetime.timedelta(**reload_interval)
        return int(time.mktime(min_full_update_date.timetuple()))

    def _calculate_min_historical_timestamp(self, config):
        history_days = config.get("history_days", 7)
        min_datetime = datetime.datetime.now() - datetime.timedelta(days=history_days)
        return int(time.mktime(min_datetime.timetuple()))