from bigpanda.agent.core.spi.plugins import Plugin
import logging
import datetime
import time
from pyzenoss import ZenossAPI
import socket

log = logging.getLogger('agent.plugins.zenoss.api')

class ZenossAPIPlugin(Plugin):

    STATUS_MAPPING = {
        2: "WARNING",
        3: "WARNING",
        4: "CRITICAL",
        5: "CRITICAL",
    }

    EVENT_STATE_MAPPING = {
        'New': None,
        'Acknowledged': None,
        'Suppressed': None,
        'Closed': 'OK',
        'Cleared': 'OK'
    }

    UNKNOWN_STATUS = "UNKNOWN"

    BIGPANDA_ACTION_NAME = "BigPanda Action"

    WARN_DESCRIPTION_LENGTH = 10 * 1000

    RETRIEVE_MAX_LIMIT = 100
    SLOW_START = 10
    OVERALL_MAX_LIMIT = RETRIEVE_MAX_LIMIT * 50 + SLOW_START

    def _init(self):
        self.zenoss_api = ZenossAPI()
        self.hostname = socket.getfqdn()

        return 0

    def _execute(self, config, state):
        self.zenoss_api.update_config(config)
        min_full_update = self._calculate_reload_interval(config)
        is_first_time = not "last_full_update" in state

        if is_first_time or state["last_full_update"] < min_full_update:
            alerts = self._update_for_open_items(is_first_time)
            state["last_full_update"] = int(time.mktime(datetime.datetime.now().timetuple()))
        else:
            min_load_time = self._calculate_min_historical_timestamp(config)
            alerts = self._retrieve_alerts(state.get("last_alert_time", min_load_time), state.get('last_alerts', list()))

        events = self._process(alerts)
        if len(events) > 0:
            state["last_alert_time"] = int(events[-1]["timestamp"])
            state['last_alerts'] = list()
            for event in events:
                if event['timestamp'] == state['last_alert_time']:
                    state['last_alerts'].append(event["properties"]["_zenoss_id"])

        if not len(events):
            log.info("Didn't find any events")
            return None, state, True
        else:
            log.info("{0} alerts found".format(len(alerts)))
            return {"alerts": events}, state, True

    def _update_for_open_items(self, is_first_time):
        log.debug("Running update_for_open_items. is the first time? " + str(is_first_time))
        return self._get_events(True)

    def _retrieve_alerts(self, last_alert_time, last_alerts):
        return self._get_events(False, last_alert_time, last_alerts)

    def _process(self, events):
        processed_events = list()

        for event in events:
            processed_event = self.process_alert(event)
            if processed_event:
                processed_events.append(processed_event)

        return processed_events

    def process_alert(self, alert):
        event = {}

        event["status"] = self._normalize_alert_status(alert.get("severity", -1), alert.get('eventState', None), alert.get('dedupid'))

        if not event["status"]:
            log.warn("Received alert with no / unrecognized status - {0}".format(alert["message"]))
            return None

        event["timestamp"] = self._to_utc_timestamp(alert.get("stateChange"))

        event["description"] = alert.get("message", None)

        event["source_system"] = "zenoss"
        event["primary_property"] = "_check"

        event["properties"] = {}
        event["properties"]["_host"] = self._normalize_host(alert.get("device", {}).get('text', None))
        component = alert.get('component', {}).get('text', None)
        event_class = alert.get("eventClass", {}).get('text', None)

        if not event_class: event_class = ''
        if not component: component = ''

        if event_class and not event_class.startswith('/'):
            event_class = '/' + event_class

        event["properties"]["_check"] = component + event_class

        if component: event['properties']['_component'] = component
        if event_class: event["properties"]["_event_class"] = event_class

        if not event['properties']['_check']: event['properties']['_check'] = 'No event'
        if not event['properties']['_host']: event['properties']['_host'] = 'No server'


        event["properties"]["_zenoss_id"] = alert.get('id', None)

        event["incident_key"] = event["properties"]["_host"] + "__" + event["properties"]["_check"]

        return event

    def _to_utc_timestamp(self, datetime):
        return int(time.mktime(time.strptime(datetime, '%Y-%m-%d %H:%M:%S')))

    def has_property(self, item, property):
        return property in item and len(item[property]) > 0

    def _normalize_host(self, host):
        if host and host.startswith('localhost.'):
            host = self.hostname
        return host

    def _normalize_alert_status(self, severity, eventState, id):
        status = None
        if eventState:
            status = self.EVENT_STATE_MAPPING.get(eventState, None)

        if not status:
            status = self.STATUS_MAPPING.get(severity, None)

        if not status:
            log.warn("Ignoring unrecognized status for: priority: {0}, value: {1}".format(severity, id))

        return status

    def _calculate_reload_interval(self, config):
        reload_interval = config.get("reload", {"days": 1})
        min_full_update_date = datetime.datetime.now() - datetime.timedelta(**reload_interval)
        return int(time.mktime(min_full_update_date.timetuple()))

    def _calculate_min_historical_timestamp(self, config):
        history_days = config.get("history_days", 7)
        min_datetime = datetime.datetime.now() - datetime.timedelta(days=history_days)
        return int(time.mktime(min_datetime.timetuple()))

    def _get_events(self, open_only, last_event_time=None, last_alerts=list()):
        # Slow start because most of the cases will retrieve only small amount of events
        events = self._get_events_recursive(list(), open_only, 0, self.SLOW_START, last_event_time, last_alerts)
        events.reverse()
        return events

    def _get_events_recursive(self, events, open_only, offset, limit, last_event_time, last_alerts):
        results = self.zenoss_api.get_events_by_start_limit(open_only, offset, limit)

        if not results.get('events'):
            return events

        for event in results.get('events'):
            stateChangeTimestamp = self._to_utc_timestamp(event.get("stateChange"))
            if stateChangeTimestamp == last_event_time:
                if event['id'] not in last_alerts:
                    events.append(event)
            elif stateChangeTimestamp > last_event_time:
                events.append(event)
            else:
                return events

        if len(events) >= self.OVERALL_MAX_LIMIT:
            return events

        return self._get_events_recursive(events, open_only, offset + limit, self.RETRIEVE_MAX_LIMIT, last_event_time, last_alerts)