import sys
import os
import logging
import traceback
import daemon
import pwd
import grp
import signal
from logging.handlers import RotatingFileHandler
from distutils.version import StrictVersion

import bigpanda
import bigpanda.agent as root
import bigpanda.agent.agent as bp_agent
import bigpanda.agent.migrations as migrations
from bigpanda.agent.core.configuration import JsonConfigurationParser, ConfigurationParseException, hash_file
from bigpanda.agent.core.client import DataApiClient


log = logging.getLogger('agent')

def setup_work_files(agent_config):
    user = agent_config.get('user', root.DEFAULT_USER)
    group = agent_config.get('group', root.DEFAULT_GROUP)
    log_file = agent_config.get('log_file', root.DEFAULT_LOG_FILE)
    pidfile = agent_config.get('pid_file', root.DEFAULT_PID_FILE)

    uid = pwd.getpwnam(user).pw_uid
    gid = grp.getgrnam(group).gr_gid

    for f in [log_file, pidfile]:
        if not os.path.exists(f):
            log.debug("%s doesn't exist, creating." % f)
            d = os.path.dirname(f)
            if not os.path.isdir(d):
                try:
                    log.debug("%s doesn't exist, creating." % d)
                    os.makedirs(d, mode=0755)
                    os.chown(d, uid, gid)
                except Exception as e:
                    raise Exception("%s: %s" % (d, e))

            try:
                open(f, 'w').close()
                os.chown(f, uid, gid)
            except Exception as e:
                raise Exception("%s: %s" % (f, e))

def write_pid(pid_file):
    pid = os.getpid()
    with open(pid_file, "w") as f:
        f.write(str(pid))

def migrate_config(config, update_file=True):
    agent_config = config.get('root', dict())
    config_version = StrictVersion(agent_config.get('version', '0.0'))
    agent_version = StrictVersion(root.VERSION)

    if agent_version > config_version:
        log.info("Migrating config from version %s to %s" % (config_version, agent_version))
        if not update_file:
            log.info("Will not write the changes back to file.")

        migrator = migrations.MigrationEngine()
        config = migrator.run_migrations(config, update_file=update_file)
    else:
        log.debug("Config file is up to date.")

    return config

def create_client(agent_config):
    backend_config = agent_config["backend"]

    return DataApiClient(backend_config["address"],
                         backend_config["api_token"],
                         backend_config["verify_certificate"],
                         agent_config["agent_id"],
                         agent_config["version"])

def check_service_disabled(config_path):
    config_dir = os.path.dirname(config_path)
    disable_service = os.path.join(config_dir, 'disable_service')
    if os.path.exists(disable_service):
        log.info("Agent is disabled, remove %s to enable." % disable_service)
        sys.exit(0)

def configure_logging(agent_config):
    log.debug("Settings up log files")
    log_file = agent_config.get('log_file', root.DEFAULT_LOG_FILE)
    log_size = agent_config.get('log_size', root.DEFAULT_LOG_SIZE)

    root_logger = logging.getLogger('')
    handler = RotatingFileHandler(log_file, maxBytes=log_size, backupCount=3)

    handler.setFormatter(bigpanda.logging_formatter_simple)

    root_logger.addHandler(handler)
    return handler.stream

def init_and_run(env, use_backup=False, detach=None):
    config_file_path, config_backup_path = root.get_config_filpath(env)
    if use_backup:
        if not os.path.exists(config_backup_path):
            sys.exit(2)
        log.info("Using backup config at %s" % config_backup_path)
        active_config_path = config_backup_path
    else:
        log.info("Using config at %s" % config_file_path)
        active_config_path = config_file_path

    check_service_disabled(active_config_path)

    # Load configuration
    try:
        log.debug("Loading configuration file from " + active_config_path)

        # We never hash the backup because we never remotely update it
        config_hash = hash_file(config_file_path)
        configurator = JsonConfigurationParser(active_config_path)
        config = configurator.parse()
    except ConfigurationParseException as e:
        log.error("Failed to parse configuration file: %s" % e)
        raise

    # Roll forward config file if necessary
    try:
        if use_backup:
            config = migrate_config(config, update_file=False)
        else:
            config = migrate_config(config, update_file=True)
    except Exception as e:
        log.exception("Failed to migrate configuration: %s" % e)
        raise

    agent_config = config["root"]

    # Setup files that need root privs to create
    try:
        setup_work_files(agent_config)
    except Exception as e:
        log.exception("Failed to create work files: %s" % e)
        raise

    # Configure logging
    try:
        logfile_stream = configure_logging(agent_config)
    except Exception as e:
        log.exception("Failed to configure logging: %s" % e)
        raise

    # Connect to servers
    try:
        client = create_client(agent_config)
    except Exception as e:
        log.exception("Failed to create backend client: %s" % e)
        raise

    # Run the agent
    try:
        # Add some useful entries to the config dict
        agent_config['config_hash'] = config_hash
        agent_config['config_file_path'] = config_file_path
        agent_config['config_backup_path'] = config_backup_path
        agent_config['using_backup'] = use_backup

        # Create and run and agent as a daemon
        user = agent_config.get('user', root.DEFAULT_USER)
        group = agent_config.get('group', root.DEFAULT_GROUP)
        uid = pwd.getpwnam(user).pw_uid
        gid = grp.getgrnam(group).gr_gid
        if detach:
            daemonize = daemon.DaemonContext(uid=uid, gid=gid, files_preserve=[logfile_stream])
            daemonize.open()
            pid_file = config.get('pid_file', root.DEFAULT_PID_FILE)
            write_pid(pid_file)

        agent = bp_agent.Agent(agent_config)
        agent.run()

        log.info("Agent is running")
    except Exception:
        message = "Failed to run Agent"
        log.exception(message)
        client.sendErrors([dict(message=message, traceback=traceback.format_exc(2))])
        raise

    # After a successful init, send the used configration to the server
    try:
        log.info("Sending configuration to server.")
        client.send('/agents/%s/config' % agent_config["agent_id"], 'POST', data=config)
    except Exception:
        message = "Failed to send configuration to server"
        log.exception(message)
        client.sendErrors([dict(message=message, traceback=traceback.format_exc(2))])

    # Keep this thread running, because otherwise signals aren't handled
    while signal.pause():
        pass
