import os
import stat


def truncated(filename, last_size, last_ino):
    stats = os.stat(filename)
    current_size = stats[stat.ST_SIZE]
    current_ino = stats[stat.ST_INO]
    return current_size < last_size or current_ino != last_ino


def changed(filename, last_size, last_ino):
    stats = os.stat(filename)
    current_size = stats[stat.ST_SIZE]
    current_ino = stats[stat.ST_INO]
    return current_size != last_size or current_ino != last_ino


class Tail:
    def __init__(self, filename, buffer_size=16384, offset=0, last_ino=None, last_size=0):
        """
        Create a new tail instance.
        Create a tail object which periodicly polls the specified file looking
        for new data which was written.  The callback routine is called for each
        new line found in the file.

        @return Nothing
        @param filename File to read.
        @param processingFun Function which takes at least one argument - the line - called with each
                line read from the file.
        @*args any additional arguments needed to be passed to the processing function
        @param tailbytes Specifies bytes from end of file to start reading
                (defaults to 0, meaning skip entire file, -1 means read full file).
        @param buffer_size Number of bytes to read in every call to #process()
        """
        self.filename = filename
        self.buffer_size = buffer_size
        self.EOF = -1

        self._set_state(offset, last_ino, last_size)

    def _set_state(self, offset, last_ino, last_size):
        self.offset = offset
        self.last_ino = last_ino
        self.last_size = last_size

    def _reset_state(self):
        self._set_state(0, None, 0)

    def get_state(self):
        return self.offset, self.last_ino, self.last_size

    def process(self):
        '''Examine file looking for new lines.
        When called, this function will process all lines in the file being
        tailed, detect the original file being renamed or reopened, etc...
        This should be called periodicly to look for activity on the file.

        @return Nothing
        '''

        try:
            fp = open(self.filename, 'r')

            stats = os.stat(self.filename)
            if not self.last_ino:
                self.last_ino = stats[stat.ST_INO]

            if self.offset > 0:
                fp.seek(self.offset)

        except Exception, e:
            self._reset_state()
            raise e

        # Check if the file was truncated since the last time we read from it
        if truncated(self.filename, self.last_size, self.last_ino):
            fp.close()
            self._reset_state()
            raise StopIteration()

        # New lines will be read only if the file has changed
        if self.buffer_size == -1:
            lines = fp.readlines()
        else:
            lines = fp.readlines(self.buffer_size)
        self.offset = fp.tell()
        if len(lines):
            for line in lines:
                yield line.strip()
        else:
            yield self.EOF

        # Update last known file-size
        stats = os.stat(self.filename)
        self.last_size = stats[stat.ST_SIZE]

        fp.close()